import json
from io import StringIO
from typing import Iterable, Optional, TextIO

import yaml
from rdflib import URIRef

from iolanta.loaders import Loader
from iolanta.models import LDContext, LDDocument, Quad
from iolanta.parsers.json import JSON

try:  # noqa
    from yaml import CSafeLoader as SafeLoader  # noqa
except ImportError:
    from yaml import SafeLoader  # type: ignore   # noqa


class YAML(JSON):
    """Load YAML data."""

    def as_jsonld_document(self, raw_data: TextIO) -> LDDocument:
        """Read YAML content and adapt it to JSON-LD format."""
        return yaml.load(raw_data, Loader=SafeLoader)

    def as_quad_stream(
        self,
        raw_data: TextIO,
        iri: Optional[URIRef],
        context: LDContext,
        root_loader: Loader,
    ) -> Iterable[Quad]:
        """Read YAML-LD data into a quad stream."""
        json_data = self.as_jsonld_document(raw_data)

        return JSON().as_quad_stream(
            raw_data=StringIO(
                json.dumps(
                    json_data,
                    ensure_ascii=False,
                    default=str,
                ),
            ),
            iri=iri,
            context=context,
            root_loader=root_loader,
        )



import logging
from pathlib import Path
from typing import Optional

from rich.console import Console
from rich.table import Table
from typer import Argument, Context, Option, Typer

from iolanta.cli.formatters.choose import cli_print
from iolanta.iolanta import Iolanta
from iolanta.models import QueryResultsFormat
from iolanta.shortcuts import construct_root_loader

logger = logging.getLogger(__name__)


def construct_app() -> Typer:
    iolanta = Iolanta(
        logger=logger,
        loader=construct_root_loader(),
    )

    cli = Typer(
        no_args_is_help=True,
        context_settings={
            'obj': iolanta,
        },
    )

    plugins = iolanta.plugins

    for plugin in plugins:
        if (subcommand := plugin.typer_app) is not None:
            cli.add_typer(subcommand)

    return cli


app = construct_app()


@app.callback()
def callback(
    context: Context,
    log_level: str = 'info',
    source: Path = Option(
        Path.cwd,
        '--from',
        help='File or directory to read data from',
        exists=True,
    ),
):
    iolanta: Iolanta = context.obj

    iolanta.logger.setLevel(
        {
            'info': logging.INFO,
            'debug': logging.DEBUG,
        }[log_level],
    )

    iolanta.add(Path(source))


@app.command(name='render')
def render_command(
    context: Context,
    url: str,
    environment: str = Option(
        'https://iolanta.tech/cli',
        '--as',
    ),
):
    """Render a given URL."""
    iolanta: Iolanta = context.obj

    if ':' not in url:
        url = f'local:{url}'

    node = iolanta.expand_qname(url)

    Console().print(
        iolanta.render(
            node=node,
            environments=[
                iolanta.expand_qname(environment),
            ],
        ),
    )


@app.command()
def namespaces(
    context: Context,
):
    """Registered namespaces."""
    iolanta: Iolanta = context.obj

    table = Table(
        'Namespace',
        'URL',
        show_header=True,
        header_style='bold magenta',
    )

    for namespace, url in iolanta.graph.namespaces():   # type: ignore
        table.add_row(namespace, url)

    Console().print(table)


@app.command()
def query(
    context: Context,
    fmt: QueryResultsFormat = Option(
        default=QueryResultsFormat.PRETTY,
        metavar='format',
    ),
    query_text: Optional[str] = Argument(
        None,
        metavar='query',
        help='SPARQL query text. Will be read from stdin if empty.',
    ),
    use_qnames: bool = Option(
        default=True,
        help='Collapse URLs into QNames.',
    ),
):
    """Query Iolanta graph with SPARQL."""
    iolanta: Iolanta = context.obj

    cli_print(
        query_result=iolanta.query(query_text),
        output_format=fmt,
        display_iri_as_qname=use_qnames,
        graph=iolanta.graph,
    )



from pathlib import Path
from typing import Dict, Iterable, Type

from rdflib import URIRef
from urlpath import URL

from iolanta.loaders.data_type_choice import DataTypeChoiceLoader
from iolanta.loaders.dict_loader import DictLoader
from iolanta.loaders.http import HTTP
from iolanta.loaders.local_directory import Loader, LocalDirectory
from iolanta.loaders.local_file import LocalFile
from iolanta.loaders.scheme_choice import SchemeChoiceLoader
from iolanta.models import LDContext, LDDocument, Quad


def choose_loader_by_url(url: URL) -> Type[Loader[URL]]:
    """Find loader by URL scheme."""
    return LocalDirectory


def as_document(path: Path) -> LDDocument:
    """Retrieve the document presented by the specified URL."""
    return LocalFile().as_jsonld_document(path)


def construct_root_loader() -> DataTypeChoiceLoader:
    # FIXME: Generalize this using endpoints
    return DataTypeChoiceLoader(
        loader_by_data_type={
            dict: DictLoader(),
            Path: LocalDirectory(),
            URL: SchemeChoiceLoader(
                loader_by_scheme={
                    'file': LocalDirectory(),
                    'http': HTTP(),
                    'https': HTTP(),
                },
            ),
        },
    )


def as_quad_stream(
    url: URL,
    iri: URIRef,
    default_context: LDContext,
    root_directory: Path,
    named_contexts: Dict[str, LDContext],
) -> Iterable[Quad]:
    """Retrieve the stream presented by the specified URL."""
    root_loader = construct_root_loader(
        default_context=default_context,
        root_directory=root_directory,
    )

    return root_loader.as_quad_stream(
        source=url,
        iri=iri,
        context=default_context,
    )



from dataclasses import dataclass
from typing import Any, Dict, Iterable, Optional, TextIO

from rdflib import URIRef
from urlpath import URL

from iolanta.loaders.base import Loader, SourceType
from iolanta.models import LDContext, LDDocument, Quad


@dataclass(frozen=True)
class DataTypeChoiceLoader(Loader[Any]):   # type: ignore
    """Try to load a file via several loaders."""

    loader_by_data_type: Dict[type, Loader[Any]]    # type: ignore

    def choose_parser_class(self, source: SourceType):
        raise ValueError('choose_parser_class')

    def as_file(self, source: SourceType) -> TextIO:
        raise ValueError('as_file')

    def find_context(self, source: SourceType) -> LDContext:
        raise ValueError('find_context')

    def resolve_loader(self, source: Any):   # type: ignore
        """Find loader instance by URL."""
        for source_type, loader in self.loader_by_data_type.items():
            if isinstance(source, source_type):
                return loader

        source_type = type(source)
        raise ValueError(
            f'Cannot find a loader for source: {source} '
            f'of type: {source_type}',
        )

    def as_jsonld_document(
        self,
        source: URL,
        iri: Optional[URIRef] = None,
    ) -> LDDocument:
        """Represent a file as a JSON-LD document."""
        return self.resolve_loader(
            source=source,
        ).as_jsonld_document(
            source=source,
            iri=iri,
        )

    def as_quad_stream(
        self,
        source: str,
        iri: Optional[URIRef],
        root_loader: Optional[Loader[URL]] = None,
        context: Optional[LDContext] = None,
    ) -> Iterable[Quad]:
        """Convert data into a stream of RDF quads."""
        return self.resolve_loader(
            source=source,
        ).as_quad_stream(
            source=source,
            iri=iri,
            root_loader=root_loader or self,
            context=context,
        )



import itertools

import funcy
from rich.table import Table

from iolanta.cli.formatters.node_to_qname import node_to_qname
from iolanta.cli.formatters.pretty import pretty_print_value
from iolanta.facet.rich import Renderable, RichFacet


class CLI(RichFacet):
    def show(self) -> Renderable:
        rows = self.query(
            '''
            SELECT ?property ?value WHERE {
                $iri ?property ?value .
            } ORDER BY ?property ?value
            ''',
            iri=self.iri,
        )

        pairs = [
            (row['property'], row['value'])
            for row in rows
        ]

        dossier = {
            property_iri: list(
                map(
                    funcy.last,
                    property_values,
                ),
            )
            for property_iri, property_values
            in itertools.groupby(
                pairs,
                key=funcy.first,
            )
        }

        table = Table.grid(padding=1, pad_edge=True)
        table.title = str(self.iri)
        table.add_column(
            "Properties",
            no_wrap=True,
            justify="left",
            style="bold blue",
        )
        table.add_column("Values")

        for property_iri, property_values in dossier.items():
            table.add_row(
                # FIXME:
                #   title: Render property instead of pretty-printing raw IRI
                #   is-blocked-by: record-render-strategy
                pretty_print_value(
                    node_to_qname(
                        property_iri,
                        self.iolanta.graph,
                    ),
                ),
                '\n'.join(
                    map(
                        str,
                        # FIXME:
                        #   title: Render property values instead of prettyprint
                        #   is-blocked-by: record-render-strategy
                        property_values,
                    ),
                ),
            )

        return table



import logging
from typing import List, Optional, Tuple, Type

from rdflib import ConjunctiveGraph

from iolanta.entry_points import entry_points
from iolanta.graph_providers.base import GraphProvider
from iolanta.graph_providers.errors import GraphCannotBeProvided

logger = logging.getLogger(__name__)


GraphProviderTuple = Tuple[str, Type[GraphProvider]]


def installed_graph_providers() -> List[GraphProviderTuple]:
    """Find installed graph providers."""
    return [
        (entry_point.name, entry_point.load())   # type: ignore
        for entry_point in entry_points(group='iolanta.graph')  # type: ignore
    ]


def construct_graph_from_installed_providers() -> Optional[ConjunctiveGraph]:
    """
    Go over installed providers.

    See if they can provide us with a graph.
    """
    for provider_name, provider_class in installed_graph_providers():
        try:
            return provider_class().retrieve_graph()
        except GraphCannotBeProvided:
            logger.info('Provider %s cannot provide a graph.', provider_name)

    return None



from functools import cached_property
from typing import Optional, cast

from more_itertools import first
from rdflib.term import URIRef

from iolanta.facet.base import FacetSearchAttempt


class FindFacetByInstance(FacetSearchAttempt):
    """
    Find facet by direct link.

    Look for such `?facet` that `{self.node}` `iolanta:facet` `?facet`, whereas
    `?facet` `iolanta:supports` `{self.environment}`.
    """

    @cached_property
    def facet(self) -> Optional[URIRef]:
        """Find facet."""
        rows = self.ldflex.query(
            '''
            SELECT ?facet WHERE {
                $node iolanta:facet ?facet .
                ?facet iolanta:supports $env .
            }
            ''',
            node=self.node,
            env=self.environment,
        )

        try:
            return cast(URIRef, first(rows)['facet'])
        except (ValueError, TypeError):
            return None



from pathlib import Path
from typing import Optional, TypedDict

import funcy
from dominate.tags import a, span  # noqa: WPS347
from rdflib import Literal

from iolanta.facet import Facet


class DefaultRow(TypedDict):
    """A few properties of the object we use as row heading."""

    label: Optional[Literal]
    symbol: Optional[Literal]
    url: Optional[Literal]
    comment: Optional[Literal]


class Default(Facet):
    """Default renderer."""

    def show(self):
        """Render the column."""
        row: DefaultRow = funcy.first(
            self.stored_query('default.sparql', iri=self.iri),
        )

        if row.get('url'):
            return self._render_link(row)

        return self._render_label(row)

    def _render_link(self, row: DefaultRow):
        comment = row.get('comment')
        return a(
            self._render_label({   # type: ignore
                row_key: row_value
                for row_key, row_value in row.items()
                if row_key != 'comment'
            }),
            href=row['url'],
            title=comment,
        )

    def _render_label(self, row: DefaultRow):
        label = row.get('label')

        if label is None:
            label = self._render_fallback()

        if symbol := row.get('symbol'):
            label = f'{symbol} {label}'

        if comment := row.get('comment'):
            return span(
                label,
                title=comment,
            )

        return label

    def _render_fallback(self) -> str:
        string_iri = str(self.iri)

        if string_iri.startswith('local:'):
            string_iri = string_iri.removeprefix(
                'local:',
            ).replace(
                '_', ' ',
            ).replace(
                '-', ' ',
            ).capitalize()

        return string_iri



from dataclasses import dataclass

from documented import DocumentedError
from rdflib import Literal
from rdflib.term import Node

from iolanta.facet import Facet
from iolanta.facets.bool_literal import NotALiteral


class CodeLiteral(Facet):
    """Render code strings."""

    def show(self):
        """Render as icon."""
        if not isinstance(self.iri, Literal):
            raise NotALiteral(
                node=self.iri,
            )

        return f'<code>{self.iri.value}</code>'



import json
from typing import Iterable, Optional, TextIO

from rdflib import URIRef
from urlpath import URL

from iolanta.loaders.base import Loader
from iolanta.models import LDContext, LDDocument, Quad
from iolanta.parsers.base import Parser
from iolanta.parsers.dict_parser import DictParser


class JSON(Parser[TextIO]):
    """Load JSON data."""

    def as_jsonld_document(self, raw_data: TextIO) -> LDDocument:
        """Read JSON content as a JSON-LD document."""
        return json.load(raw_data)

    def as_quad_stream(
        self,
        raw_data: TextIO,
        iri: Optional[URIRef],
        context: LDContext,
        root_loader: Loader[URL],
    ) -> Iterable[Quad]:
        """Read JSON-LD data into a quad stream."""
        document = self.as_jsonld_document(raw_data)

        return DictParser().as_quad_stream(
            raw_data=document,
            iri=iri,
            context=context,
            root_loader=root_loader,
        )





from dataclasses import dataclass
from functools import cached_property
from typing import Optional

from documented import Documented
from rdflib.term import Node, URIRef

from ldflex import LDFlex


@dataclass
class FacetSearchAttempt(Documented):
    """Find facet for a node."""

    ldflex: LDFlex
    node: Node
    environment: URIRef

    @cached_property
    def facet(self) -> Optional[URIRef]:
        """Find facet."""
        raise NotImplementedError()

    def __bool__(self):
        """Find if facet was found or not."""
        return self.facet is not None



from abc import ABC, abstractmethod

import rdflib


class GraphProvider(ABC):
    """
    Graph provider.

    Out of the blue, a graph provider returns a graph that iolanta is going to
    work off of.
    """

    @abstractmethod
    def retrieve_graph(self) -> rdflib.Dataset:
        """Find the graph."""



from dataclasses import dataclass

from documented import DocumentedError
from rdflib import Literal
from rdflib.term import Node

from iolanta.facet import Facet


@dataclass
class NotALiteral(DocumentedError):
    """
    Node `{self.node}` is not a literal.

    It is in fact a `{self.node_type}`. `BoolLiteral` facet only supports RDF
    literal objects.
    """

    node: Node

    @property
    def node_type(self):
        """Node type name."""
        return self.node.__class__.__name__


class BoolLiteral(Facet):
    """Render bool values."""

    def show(self):
        """Render as icon."""
        if not isinstance(self.iri, Literal):
            raise NotALiteral(
                node=self.iri,
            )

        return '✔️' if self.iri.value else '❌'



from abc import ABC
from dataclasses import dataclass
from typing import (
    Any,
    Dict,
    Generic,
    Iterable,
    Optional,
    TextIO,
    TypedDict,
    TypeVar,
)

from rdflib import URIRef
from urlpath import URL

from iolanta.conversions import url_to_iri
from iolanta.ensure_is_context import ensure_is_context
from iolanta.models import LDContext, LDDocument, Quad
from iolanta.namespaces import PYTHON

SourceType = TypeVar('SourceType')


PyLDOptions = Dict[str, Any]   # type: ignore

PyLDResponse = TypedDict(
    'PyLDResponse', {
        'contentType': str,
        'contextUrl': Optional[str],
        'documentUrl': str,
        'document': LDDocument,
    },
)


def term_for_python_class(cls: type) -> URIRef:
    """Construct term for Python class."""
    return PYTHON.term(f'{cls.__module__}.{cls.__qualname__}')


# noinspection TaskProblemsInspection
@dataclass(frozen=True)
class Loader(ABC, Generic[SourceType]):
    """
    Base class for loaders.

    Loader receives a URL (or a path) to certain location. It is responsible for
    reading data from that location and returning it as a stream of RDF quads.

    Usually, depending on the data format, Loader leverages Parsers for that
    purpose.
    """

    @classmethod
    def loader_class_iri(cls) -> URIRef:
        """Import path to the loader class."""
        return term_for_python_class(cls)

    def choose_parser_class(self, source: SourceType):
        """Find which parser class to use for this URL."""
        raise NotImplementedError(
            f'{self}.choose_parser_class() is not implemented.',
        )

    def as_jsonld_document(
        self,
        source: SourceType,
        iri: Optional[URIRef] = None,
    ) -> LDDocument:
        """Represent a file as a JSON-LD document."""
        raise NotImplementedError(
            f'{self}.as_jsonld_document() is not implemented.',
        )

    def as_file(self, source: SourceType) -> TextIO:
        """Construct a file-like object."""
        raise NotImplementedError()

    def as_quad_stream(
        self,
        source: SourceType,
        iri: Optional[URIRef],
        root_loader: 'Loader[SourceType]',
        context: Optional[LDContext] = None,
    ) -> Iterable[Quad]:
        """Convert data into a stream of RDF quads."""
        raise NotImplementedError(
            f'{self}.as_quad_stream() is not implemented.',
        )

    def find_context(self, source: SourceType) -> LDContext:
        """Find context for the file."""
        raise NotImplementedError(
            f'{self}.find_context() is not implemented.',
        )

    def __call__(self, source: str, options: PyLDOptions) -> PyLDResponse:
        """
        Call the loader to retrieve the document in a PYLD friendly format.

        Used to resolve remote contexts.

        The type of `source` parameter is intentionally `str`: that's the only
        thing which pyld can do.
        """
        source = URL(source)

        document = ensure_is_context(
            self.as_jsonld_document(
                source=URL(source),
                iri=url_to_iri(source),
            ),
        )

        return {
            'document': document,
            'contextUrl': None,
            'documentUrl': source,
            'contentType': 'application/ld+json',
        }



from dataclasses import dataclass
from functools import cached_property
from pathlib import Path
from typing import Optional

import funcy
from more_itertools import first
from rdflib.term import URIRef

from iolanta.facet.base import FacetSearchAttempt
from iolanta.facet.errors import MultipleFacetsFoundForTypes

QUERY = (
    Path(__file__).parent / 'sparql/find_facet_for_literal.sparql'
).read_text()


@dataclass
class FindFacetForLiteral(FacetSearchAttempt):
    """Find facet by node type."""

    @cached_property
    def facet(self) -> Optional[URIRef]:
        """Find facet."""
        rows = self.ldflex.query(
            QUERY,
            env=self.environment,
        )

        facets = list(funcy.pluck('facet', rows))

        if len(facets) > 1:
            raise MultipleFacetsFoundForTypes(
                node=self.node,
                facets=facets,
            )

        return first(facets, None)



import functools
import logging
from dataclasses import dataclass, field
from functools import cached_property
from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional, Type, Union

import owlrl
from owlrl import OWLRL_Extension
from rdflib import ConjunctiveGraph, Namespace, URIRef
from rdflib.term import Node

from iolanta import entry_points
from iolanta.loaders import Loader
from iolanta.loaders.base import SourceType
from iolanta.loaders.local_directory import merge_contexts
from iolanta.models import LDContext, NotLiteralNode
from iolanta.namespaces import IOLANTA, LOCAL
from iolanta.parsers.yaml import YAML
from iolanta.plugin import Plugin
from iolanta.shortcuts import construct_root_loader
from ldflex import LDFlex


@dataclass
class Iolanta:
    """Iolanta is a Semantic web browser."""

    loader: Loader[SourceType] = field(default_factory=construct_root_loader)
    graph: ConjunctiveGraph = field(
        default_factory=functools.partial(
            ConjunctiveGraph,
            identifier=LOCAL.term('_inference'),
        ),
    )
    force_plugins: List[Type[Plugin]] = field(default_factory=list)

    logger: logging.Logger = field(
        default_factory=functools.partial(
            logging.getLogger,
            name='iolanta',
        ),
    )

    @property
    def plugin_classes(self) -> List[Type[Plugin]]:
        """Installed Iolanta plugins."""
        return self.force_plugins or entry_points.plugins('iolanta.plugins')

    @cached_property
    def plugins(self) -> List[Plugin]:
        return [
            plugin_class(iolanta=self)
            for plugin_class in self.plugin_classes
        ]

    @cached_property
    def ldflex(self) -> LDFlex:
        """LDFlex is a wrapper to make SPARQL querying RDF graphs bearable."""
        return LDFlex(self.graph)

    @cached_property
    def namespaces_to_bind(self) -> Dict[str, Namespace]:
        return {
            key: Namespace(value)
            for key, value in self.default_context['@context'].items()
            if (
                isinstance(value, str)
                and not value.startswith('@')
                and not key.startswith('@')
            )
        }

    def add(  # type: ignore
        self,
        source: Any,
        context: Optional[LDContext] = None,
        graph_iri: Optional[URIRef] = None,
    ) -> 'Iolanta':
        """Parse & load information from given URL into the graph."""
        quads = list(
            self.loader.as_quad_stream(
                source=source,
                iri=graph_iri,
                context=context or self.default_context,
                root_loader=self.loader,
            ),
        )

        self.graph.addN(quads)

        self.bind_namespaces(**self.namespaces_to_bind)

        self.infer()

        return self

    def infer(self) -> 'Iolanta':
        self.logger.info('Inference: OWL RL started...')
        owlrl.DeductiveClosure(OWLRL_Extension).expand(self.graph)
        self.logger.info('Inference: OWL RL complete.')

        return self

    def bind_namespaces(self, **mappings: Namespace) -> 'Iolanta':
        """Bind namespaces."""
        self.graph.bind(prefix='local', namespace=LOCAL)

        for prefix, namespace in mappings.items():
            self.graph.bind(prefix=prefix, namespace=namespace)

        return self

    @property
    def query(self):
        return self.ldflex.query

    @cached_property
    def context_paths(self) -> Iterable[Path]:
        directory = Path(__file__).parent / 'data'

        yield directory / 'context.yaml'

        for plugin in self.plugins:
            if path := plugin.context_path:
                yield path

    @cached_property
    def default_context(self) -> LDContext:
        """Construct default context from plugins."""
        context_documents = [
            YAML().as_jsonld_document(path.open('r'))
            for path in self.context_paths
        ]

        for context in context_documents:
            if isinstance(context, list):
                raise ValueError('Context cannot be a list: %s', context)

        return merge_contexts(*context_documents)   # type: ignore

    def add_files_from_plugins(self):
        for plugin in self.plugins:
            self.add(plugin.data_files)

    def __post_init__(self):
        self.add_files_from_plugins()

    def expand_qname(self, qname: str) -> URIRef:
        try:
            return self.graph.namespace_manager.expand_curie(qname)
        except ValueError:
            return URIRef(qname)

    def render(
        self,
        node: Union[str, Node],
        environments: Optional[Union[str, List[NotLiteralNode]]] = None,
    ) -> Any:
        """Find an Iolanta facet for a node and render it."""
        # FIXME: Convert to a global import
        from iolanta.facet.errors import FacetError, FacetNotFound
        from iolanta.renderer import Render, render_facet, resolve_facet

        if isinstance(environments, str):
            environments = [environments]

        if not environments:
            environments = [IOLANTA.html]

        self.logger.debug('Environments: %s', environments)

        facet_search_attempt = Render(
            ldflex=self.ldflex,
        ).search_for_facet(
            node=node,
            environments=environments,
        )

        facet_class = resolve_facet(
            iri=facet_search_attempt.facet,
        )

        facet = facet_class(
            iri=node,
            iolanta=self,
            environment=facet_search_attempt.environment,
        )

        try:
            return render_facet(
                node=node,
                facet=facet,
                environments=environments,
                debug_mode=False,  # iolanta.is_debug_mode(node),
            )

        except (FacetError, FacetNotFound):  # noqa: WPS329
            # Prevent nested `FacetError` exceptions.
            raise

        except Exception as err:
            raise FacetError(
                node=node,
                facet_iri=facet_search_attempt.facet,
                facet_search_attempt=facet_search_attempt,
                error=err,
            ) from err



from abc import ABC
from typing import Any, Protocol, Union

from iolanta.facet import Facet


class Rich(Protocol):
    __rich__: Any  # type: ignore


Renderable = Union[Rich, str]


class RichFacet(ABC, Facet[Renderable]):
    """Render stuff in console with Rich."""



from typing import Optional

from rdflib import Graph

from iolanta.cli.formatters.csv import csv_print
from iolanta.cli.formatters.json import print_json
from iolanta.cli.formatters.node_to_qname import node_to_qname
from iolanta.cli.formatters.pretty import pretty_print
from iolanta.models import QueryResultsFormat
from ldflex.ldflex import SelectResult


# @cli_print.instance(SelectResult)
def cli_print(
    query_result: SelectResult,
    output_format: QueryResultsFormat,
    display_iri_as_qname: bool = True,
    graph: Optional[Graph] = None,
):
    if display_iri_as_qname:
        if graph is None:
            raise NotImplementedError(
                'Cannot compute QNames if graph is not provided.',
            )

        query_result = SelectResult([
            {
                key: node_to_qname(node, graph)
                for key, node in row.items()
            }
            for row in query_result
        ])

    {
        QueryResultsFormat.CSV: csv_print,
        QueryResultsFormat.PRETTY: pretty_print,
        QueryResultsFormat.JSON: print_json,
    }[output_format](query_result)   # type: ignore



from dataclasses import dataclass
from pathlib import Path
from typing import Iterable, Optional, TextIO, Type

from rdflib import Literal, URIRef

from iolanta.conversions import url_to_iri
from iolanta.loaders.base import Loader
from iolanta.loaders.errors import IsAContext, ParserNotFound
from iolanta.models import LDContext, LDDocument, Quad
from iolanta.namespaces import IOLANTA
from iolanta.parsers.base import Parser
from iolanta.parsers.json import JSON
from iolanta.parsers.markdown import Markdown
from iolanta.parsers.yaml import YAML


def choose_parser_by_extension(path: Path) -> Type[Parser]:
    """
    Choose parser class based on file extension.

    FIXME this is currently hard coded; need to change to a more extensible
      mechanism.
    """
    try:
        return {
            '.yaml': YAML,
            '.json': JSON,
            '.md': Markdown,
        }[path.suffix]
    except KeyError:
        raise ParserNotFound(path=path)


@dataclass(frozen=True)
class LocalFile(Loader[Path]):
    """
    Retrieve Linked Data from a file on local disk.

    Requires Path with file:// scheme as input.
    """

    def find_context(self, source: str) -> LDContext:
        return {}

    def choose_parser_class(self, source: Path) -> Type[Parser]:
        return choose_parser_by_extension(source)

    def as_quad_stream(
        self,
        source: Path,
        root_loader: Loader[Path],
        iri: Optional[URIRef] = None,
        context: Optional[LDContext] = None,
    ) -> Iterable[Quad]:
        """Extract a sequence of quads from a local file."""
        if source.stem == 'context':
            raise IsAContext(path=source)

        try:
            parser_class = self.choose_parser_class(source)
        except ParserNotFound:
            return []

        if iri is None:
            iri = url_to_iri(source)

        with source.open() as text_io:
            yield from parser_class().as_quad_stream(
                raw_data=text_io,
                iri=iri,
                context=context,
                root_loader=root_loader,
            )

            yield Quad(
                iri,
                IOLANTA.fileName,
                Literal(source.name),
                URIRef('https://iolanta.tech/loaders/local-file'),
            )

    def as_file(self, source: Path) -> TextIO:
        """Construct a file-like object."""
        with source.open() as text_io:
            return text_io

    def as_jsonld_document(
        self,
        source: Path,
        iri: Optional[URIRef] = None,
    ) -> LDDocument:
        """As JSON-LD document."""
        parser_class: Type[Parser] = self.choose_parser_class(source)
        with source.open() as text_io:
            document = parser_class().as_jsonld_document(text_io)

        if iri is not None and isinstance(document, dict):
            document.setdefault('@id', str(iri))

        return document



import dataclasses
from dataclasses import dataclass, field
from functools import reduce
from pathlib import Path
from typing import Iterable, List, Optional, TextIO, Type

from rdflib import URIRef

from iolanta.context import merge
from iolanta.conversions import path_to_iri
from iolanta.ensure_is_context import NotAContext, ensure_is_context
from iolanta.loaders.base import Loader, SourceType
from iolanta.loaders.local_file import LocalFile
from iolanta.models import LDContext, LDDocument, Quad
from iolanta.namespaces import IOLANTA
from iolanta.parsers.base import Parser


def merge_contexts(*contexts: LDContext) -> LDContext:
    return reduce(
        merge,
        filter(bool, contexts),
        {},
    )


@dataclass(frozen=True)
class LocalDirectory(Loader[Path]):
    """
    Retrieve Linked Data from a file on local disk.

    Requires Path with file:// scheme as input.
    """

    context_filenames: List[str] = field(
        default_factory=lambda: [
            'context.yaml',
            'context.yml',
            'context.json',
        ],
    )

    def find_context(self, source: SourceType) -> LDContext:
        raise ValueError('?!!!???')

    def directory_level_context(self, path: Path) -> Optional[LDContext]:
        for file_name in self.context_filenames:
            if (context_path := path / file_name).is_file():
                document = LocalFile().as_jsonld_document(
                    source=context_path,
                )

                if document:
                    try:
                        return ensure_is_context(document)
                    except NotAContext as err:
                        raise dataclasses.replace(
                            err,
                            path=context_path,
                        )
        return None

    def choose_parser_class(self, source: Path) -> Type[Parser]:
        """Choose parser class based on file extension."""
        raise ValueError('This is a directory')

    def as_quad_stream(
        self,
        source: Path,
        iri: Optional[URIRef],
        root_loader: Loader[Path],
        context: Optional[LDContext] = None,
    ) -> Iterable[Quad]:
        """Extract a sequence of quads from a local file."""
        if iri is None:
            iri = path_to_iri(source.absolute())

        if not source.is_dir():
            yield from LocalFile().as_quad_stream(
                source=source,
                root_loader=root_loader,
                iri=iri,
                context=context,
            )
            return

        context = merge_contexts(
            context,
            self.directory_level_context(source),
        )

        for child in source.iterdir():
            if not iri.endswith('/'):
                iri = URIRef(f'{iri}/')

            child_iri = URIRef(f'{iri}{child.name}')

            if child.is_dir():
                child_iri += '/'

                yield from LocalDirectory().as_quad_stream(
                    source=child,
                    iri=child_iri,
                    root_loader=root_loader,
                    context=context,
                )

            elif child.stem != 'context':
                yield from LocalFile().as_quad_stream(
                    source=child,
                    iri=child_iri,
                    root_loader=root_loader,
                    context=context,
                )

            if iri is not None:
                yield Quad(
                    subject=child_iri,
                    predicate=IOLANTA.isChildOf,
                    object=iri,
                    graph=URIRef(
                        'https://iolanta.tech/loaders/local-directory',
                    ),
                )

    def as_file(self, source: Path) -> TextIO:
        """Construct a file-like object."""
        with source.open() as text_io:
            return text_io

    def as_jsonld_document(
        self,
        source: Path,
        iri: Optional[URIRef] = None,
    ) -> LDDocument:
        """As JSON-LD document."""
        raise ValueError('This is a directory.')



import inspect
from abc import ABC
from dataclasses import dataclass, field
from pathlib import Path
from typing import Optional

from typer import Typer


@dataclass
class Plugin(ABC):
    """Base Iolanta plugin."""

    iolanta: 'iolanta.Iolanta' = field(repr=False)

    @property
    def typer_app(self) -> Optional[Typer]:
        """Typer app for this plugin's CLI."""
        return None

    @property
    def files_directory(self) -> Path:
        return Path(inspect.getfile(self.__class__)).parent / 'data'

    @property
    def context_path(self) -> Optional[Path]:
        if (context_path := self.files_directory / 'context.yaml').is_file():
            return context_path

    @property
    def data_files(self):
        return self.files_directory



from copy import deepcopy

from deepmerge import always_merger

from iolanta.models import LDContext


def merge(first: LDContext, second: LDContext) -> LDContext:
    """
    Merge two contexts into one.

    Second context can override the first.
    """
    if isinstance(second, list):
        return [
            merge(first, sub_second)
            for sub_second in second
        ]

    return deepcopy(
        always_merger.merge(
            base=first,
            nxt=second,
        ),
    )



import inspect
from dataclasses import dataclass
from functools import cached_property
from pathlib import Path
from typing import Any, Generic, List, Optional, TypeVar, Union

from rdflib.term import BNode, Node, URIRef

from iolanta.iolanta import Iolanta
from iolanta.models import NotLiteralNode
from ldflex import LDFlex
from ldflex.ldflex import QueryResult, SPARQLQueryArgument

FacetOutput = TypeVar('FacetOutput')


@dataclass
class Facet(Generic[FacetOutput]):
    """Base facet class."""

    iri: NotLiteralNode
    iolanta: Iolanta
    environment: Optional[URIRef] = None

    @property
    def stored_queries_path(self) -> Path:
        return Path(inspect.getfile(self.__class__)).parent / 'sparql'

    @property
    def ldflex(self) -> LDFlex:
        """Extract LDFLex instance."""
        return self.iolanta.ldflex

    @cached_property
    def uriref(self) -> NotLiteralNode:
        """Format as URIRef."""
        if isinstance(self.iri, BNode):
            return self.iri

        return URIRef(self.iri)

    def query(
        self,
        query_text: str,
        **kwargs: SPARQLQueryArgument,
    ) -> QueryResult:
        """SPARQL query."""
        return self.ldflex.query(
            query_text=query_text,
            **kwargs,
        )

    def render(
        self,
        node: Union[str, Node],
        environments: Optional[Union[str, List[NotLiteralNode]]] = None,
    ) -> Any:
        """Shortcut to render something via iolanta."""
        return self.iolanta.render(
            node=node,
            environments=environments,
        )

    def stored_query(self, file_name: str, **kwargs: SPARQLQueryArgument):
        """Execute a stored SPARQL query."""
        query_text = (self.stored_queries_path / file_name).read_text()
        return self.query(
            query_text=query_text,
            **kwargs,
        )

    def show(self) -> FacetOutput:
        """Render the facet."""
        raise NotImplementedError()

    @property
    def language(self):
        # return self.iolanta.language
        return 'en'

    def __str__(self):
        """Render."""
        return str(self.show())



import json
from io import StringIO
from typing import Iterable, Optional, TextIO

import frontmatter
from rdflib import URIRef
from yaml.parser import ParserError
from yaml.scanner import ScannerError

from iolanta.loaders import Loader
from iolanta.loaders.base import SourceType
from iolanta.models import LDContext, LDDocument, Quad
from iolanta.parsers.errors import YAMLError
from iolanta.parsers.json import JSON
from iolanta.parsers.yaml import YAML

try:  # noqa
    from yaml import CSafeLoader as SafeLoader  # noqa
except ImportError:
    from yaml import SafeLoader  # type: ignore   # noqa


class Markdown(YAML):
    """Load YAML data."""

    def as_jsonld_document(self, raw_data: TextIO) -> LDDocument:
        """Read YAML content and adapt it to JSON-LD format."""
        raw_data.seek(0)
        return frontmatter.load(raw_data).metadata

    def as_quad_stream(
        self,
        raw_data: TextIO,
        iri: Optional[URIRef],
        context: LDContext,
        root_loader: Loader[SourceType],
    ) -> Iterable[Quad]:
        """Assign iolanta:url and generate quad stream."""
        try:
            json_data = self.as_jsonld_document(raw_data)
        except (ScannerError, ParserError) as err:
            raise YAMLError(
                iri=iri,
                error=err,
            ) from err

        return JSON().as_quad_stream(
            raw_data=StringIO(
                json.dumps(
                    json_data,
                    ensure_ascii=False,
                    default=str,
                ),
            ),
            iri=iri,
            context=context,
            root_loader=root_loader,
        )



"""
Iolanta facet management.

This module contains a few functions which later will be refactored into
Iolanta - the generic metaverse/cyberspace browser.
"""
import pydoc
from dataclasses import dataclass
from typing import Iterable, List, Optional, Type, Union, cast

from dominate.tags import b, font, table, td, tr
from rdflib.term import Literal, Node, URIRef

from iolanta.facet.base import FacetSearchAttempt
from iolanta.facet.by_environment import FindFacetByEnvironment
from iolanta.facet.by_instance import FindFacetByInstance
from iolanta.facet.by_literal_datatype import FindFacetByLiteralDatatype
from iolanta.facet.by_type import FindFacetByInstanceTypes
from iolanta.facet.errors import FacetNotCallable, FacetNotFound
from iolanta.facet.for_literal import FindFacetForLiteral
from ldflex import LDFlex

HTML = URIRef('https://html.spec.whatwg.org/')


def resolve_facet(iri: URIRef) -> Type['iolanta.Facet']:
    """Resolve a path to a Python object to that object."""
    url = str(iri)

    if not url.startswith('python://'):
        raise Exception(
            'Octadocs only supports facets which are importable Python '
            'callables. The URLs of such facets must start with `python://`, '
            'which {url} does not comply to.'.format(
                url=url,
            ),
        )

    # It is impossible to use `urlpath` for this operation because it (or,
    # rather, one of upper classes from `urllib` that `urlpath` depends upon)
    # will lowercase the URL when parsing it - which means, irreversibly. We
    # have to resort to plain string manipulation.
    import_path = url.replace('python://', '').strip('/')

    facet = cast(Type['iolanta.Facet'], pydoc.locate(import_path))

    if not callable(facet):
        raise FacetNotCallable(
            path=import_path,
            facet=facet,
        )

    return facet


def render_facet(
    node: Node,
    facet: 'iolanta.Facet',
    environments: Optional[List[URIRef]] = None,
    debug_mode: bool = False,
):
    if debug_mode:
        return debug_node(
            node=node,
            environments=environments,
        )

    return facet.show()


def debug_node(node: Union[str, Node], environments: Optional[List[URIRef]]):
    return table(
        tr(
            td(
                b(font('Debug', color='red')),
                colspan=2,
                border=0,
            ),
        ),
        tr(
            td(b(font('Node', color='red')), border=0, align='left'),
            td(str(node), border=0, align='left'),
        ),
        tr(
            td(b(font('Environments', color='red')), border=0, align='left'),
            td(
                ', '.join(
                    map(str, environments),
                ), border=0, align='left',
            ),
        ),
        border=1,
    )


@dataclass
class Render:
    """Facet renderer."""

    ldflex: LDFlex

    def search_for_facet(
        self,
        node: Node,
        environments: List[URIRef],
    ) -> FacetSearchAttempt:
        """Find facet IRI for given node for all environments given."""
        facet_search_attempts = list(
            self.attempt_search_for_facet(
                node=node,
                environments=environments,
            ),
        )

        *failures, success = facet_search_attempts

        if success:
            return success

        raise FacetNotFound(
            node=node,
            environments=environments,
            facet_search_attempts=facet_search_attempts,
        )

    def find_facet_iri(
        self,
        node: Node,
        environments: List[URIRef],
    ) -> URIRef:
        """Find facet IRI for given node for all environments given."""
        return self.search_for_facet(
            node=node,
            environments=environments,
        ).facet

    def attempt_search_for_facet(
        self,
        node: Node,
        environments: List[URIRef],
    ) -> Iterable[FacetSearchAttempt]:
        """
        Stream of attempts to find a facet.

        If this function yields a sequence of N elements, first (N - 1) of those
        are guaranteed to be failures. The remaining one element can be both
        a failure (if nothing was found) or a success (if a facet was found).
        """
        for environment in environments:
            facet_searches = self.find_facet_iri_per_environment(
                node=node,
                environment=environment,
            )

            for facet_search in facet_searches:
                yield facet_search

                if facet_search:
                    return

    def find_facet_iri_per_environment(
        self,
        node: Node,
        environment: URIRef,
    ):
        """Find facet IRI for given node in given env."""
        if isinstance(node, Literal):
            yield from self.find_facet_iri_for_literal(
                literal=node,
                environment=environment,
            )

        yield FindFacetByInstance(
            ldflex=self.ldflex,
            node=node,
            environment=environment,
        )

        yield FindFacetByInstanceTypes(
            ldflex=self.ldflex,
            node=node,
            environment=environment,
        )

        yield FindFacetByEnvironment(
            ldflex=self.ldflex,
            node=node,
            environment=environment,
        )

    def __call__(
        self,
        node: Node,
        environment: URIRef,
    ) -> str:
        ...

    def find_facet_iri_for_literal(self, literal: Literal, environment: URIRef):
        """Find facet IRI for a literal."""
        if literal.datatype is not None:
            yield FindFacetByLiteralDatatype(
                ldflex=self.ldflex,
                node=literal,
                environment=environment,
            )

        yield FindFacetForLiteral(
            ldflex=self.ldflex,
            node=literal,
            environment=environment,
        )

        yield FindFacetByEnvironment(
            ldflex=self.ldflex,
            node=literal,
            environment=environment,
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['iolanta',
 'iolanta.cli',
 'iolanta.cli.formatters',
 'iolanta.facet',
 'iolanta.facets',
 'iolanta.graph_providers',
 'iolanta.loaders',
 'iolanta.parsers',
 'iolanta_record',
 'iolanta_record.facets',
 'ldflex']

package_data = \
{'': ['*'],
 'iolanta': ['data/*'],
 'iolanta.facet': ['sparql/*'],
 'iolanta.facets': ['sparql/*'],
 'iolanta_record': ['data/*']}

install_requires = \
['PyLD>=2.0.3,<3.0.0',
 'classes>=0.4.0,<0.5.0',
 'deepmerge>=0.1.1,<0.2.0',
 'documented>=0.1.1,<0.2.0',
 'dominate>=2.6.0,<3.0.0',
 'funcy>=1.17,<2.0',
 'more-itertools>=9.0.0,<10.0.0',
 'owlrl>=6.0.2,<7.0.0',
 'python-frontmatter>=0.5.0,<0.6.0',
 'rdflib>=6.2.0,<7.0.0',
 'requests>=2.25.1,<3.0.0',
 'rich>=13.3.1,<14.0.0',
 'typer>=0.7.0,<0.8.0',
 'urlpath>=1.1.7,<2.0.0']

entry_points = \
{'console_scripts': ['iolanta = iolanta.cli:app'],
 'iolanta.plugins': ['base = iolanta:IolantaBase',
                     'record = iolanta_record:IolantaRecord']}

setup_kwargs = {
    'name': 'iolanta',
    'version': '1.0.0',
    'description': 'Semantic Web browser',
    'long_description': '# iolanta\n\nStub repo for the iolanta browser.\n\n',
    'author': 'Anatoly Scherbakov',
    'author_email': 'altaisoft@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



from dataclasses import dataclass

from documented import DocumentedError

from iolanta.graph_providers.base import GraphProvider


@dataclass
class GraphCannotBeProvided(DocumentedError):
    """
    Graph cannot be provided.

        - Provider: {self.provider}
        - Error: {self.error}
    """

    provider: GraphProvider
    error: Exception



import dataclasses
import itertools
import json
import uuid
from dataclasses import dataclass
from typing import Any, Iterable, Optional

from documented import DocumentedError
from pyld import jsonld
from pyld.jsonld import JsonLdError, _resolved_context_cache
from rdflib import RDF, XSD, BNode, Literal, URIRef
from rdflib.term import Node
from urlpath import URL

from iolanta.loaders import Loader
from iolanta.models import LDContext, LDDocument, NotLiteralNode, Quad
from iolanta.namespaces import IOLANTA, LOCAL
from iolanta.parsers.base import Parser, RawDataType
from iolanta.parsers.errors import SpaceInProperty


class DictParser(Parser[LDDocument]):
    def as_jsonld_document(self, raw_data: LDDocument) -> LDDocument:
        return raw_data

    def as_quad_stream(
        self,
        raw_data: RawDataType,
        iri: Optional[NotLiteralNode],
        context: LDContext,
        root_loader: Loader,
    ) -> Iterable[Quad]:
        # This helps avoid weird bugs when loading data.
        _resolved_context_cache.clear()

        document = raw_data

        if iri is None:
            uid = uuid.uuid4().hex
            iri = BNode(f'_:dict:{uid}')

        document = assign_key_if_not_present(
            document=document,
            key='iolanta:subjectOf',
            default_value={
                '$id': str(iri),
            },
        )

        try:
            document = jsonld.expand(
                document,
                options={
                    'expandContext': context,
                    'documentLoader': root_loader,

                    # Explanation:
                    #   https://github.com/digitalbazaar/pyld/issues/143
                    'base': str(LOCAL),
                },
            )
        except (JsonLdError, KeyError, TypeError) as err:
            raise ExpandError(
                message=str(err),
                document=document,
                context=context,
                iri=iri,
                document_loader=root_loader,
            ) from err

        document = jsonld.flatten(document)

        static_quads = [
            Quad(iri, RDF.type, IOLANTA.File, iri),
        ]

        try:
            parsed_quads = list(
                parse_quads(
                    quads_document=jsonld.to_rdf(document),
                    # FIXME:
                    #   title: Can iri be None in a parser?
                    #   description: |
                    #     Does it make sense? If not, just change
                    #     the annotation.
                    graph=iri,  # type: ignore
                    blank_node_prefix=str(iri),
                ),
            )
        except UnresolvedIRI as err:
            raise dataclasses.replace(
                err,
                context=context,
                iri=iri,
            )

        return list(
            itertools.chain(
                parsed_quads,
                static_quads,
            ),
        )


def assign_key_if_not_present(  # type: ignore
    document: LDDocument,
    key: str,
    default_value: Any,
) -> LDDocument:
    """Add key to document if it does not exist yet."""
    if isinstance(document, dict):
        if document.get(key) is None:
            return {
                key: default_value,
                **document,
            }

        return document

    elif isinstance(document, list):
        return [
            assign_key_if_not_present(    # type: ignore
                document=sub_document,
                key=key,
                default_value=default_value,
            )
            for sub_document in document
        ]

    return document


@dataclass
class UnresolvedIRI(DocumentedError):
    """
    An unresolved IRI found.

        IRI: {self.iri}
        file: {self.file}
        prefix: {self.prefix}

    Perhaps you forgot to import appropriate context? For example:

    ```yaml
    "@context":
        - {self.prefix}: https://example.com/{self.prefix}/
    ```

    Context: {self.context}
    """

    iri: str
    prefix: str
    file: Optional[str] = None
    context: Optional[LDContext] = None


def raise_if_term_is_qname(term_value: str):
    """Raise an error if a QName is provided instead of a full IRI."""
    prefix, etc = term_value.split(':', 1)

    if etc.startswith('/'):
        return

    if prefix in {'local', 'templates'}:
        return

    raise UnresolvedIRI(
        iri=term_value,
        prefix=prefix,
    )


def parse_term(
    term,
    blank_node_prefix,
) -> Node:
    """Parse N-Quads term into a Quad."""
    if term is None:
        raise SpaceInProperty()

    term_type = term['type']
    term_value = term['value']

    if term_type == 'IRI':
        raise_if_term_is_qname(term_value)
        return URIRef(term_value)

    if term_type == 'literal':
        language = term.get('language')

        if datatype := term.get('datatype'):
            datatype = URIRef(datatype)

            # XSD.string does not provide any extra information. Removing it.
            if datatype == XSD.string:
                datatype = None

        if language and datatype:
            datatype = None

        return Literal(
            term_value,
            datatype=datatype,
            lang=language,
        )

    if term_type == 'blank node':
        return BNode(
            value=term_value.replace('_:', f'{blank_node_prefix}/'),
        )

    raise ValueError(f'Unknown term: {term}')


def parse_quads(
    quads_document,
    graph: URIRef,
    blank_node_prefix: str = '',
) -> Iterable[Quad]:
    """Parse an N-Quads output into a Quads stream."""
    for graph_name, quads in quads_document.items():
        if graph_name == '@default':
            graph_name = graph

        else:
            graph_name = URIRef(graph_name)

        for quad in quads:
            try:
                yield Quad(
                    subject=parse_term(quad['subject'], blank_node_prefix),
                    predicate=parse_term(quad['predicate'], blank_node_prefix),
                    object=parse_term(quad['object'], blank_node_prefix),
                    graph=graph_name,
                )
            except SpaceInProperty as err:
                raise dataclasses.replace(
                    err,
                    iri=graph,
                )


@dataclass
class ExpandError(DocumentedError):
    """
    JSON-LD expand operation failed.

    IRI: {self.iri}

    Context: {self.formatted_context}

    Document: {self.formatted_data}

    Error: {self.message}

    Document Loader: {self.document_loader}
    """

    message: str
    document: LDDocument
    context: LDContext
    iri: Optional[URIRef]
    document_loader: Loader[URL]

    @property
    def formatted_data(self) -> str:
        """Format document for printing."""
        return json.dumps(self.document, indent=2, ensure_ascii=False)

    @property
    def formatted_context(self):
        """Format context for printing."""
        return json.dumps(self.context, indent=2, ensure_ascii=False)



from dataclasses import dataclass
from typing import Dict, Iterable, Optional

from rdflib import URIRef
from urlpath import URL

from iolanta.conversions import url_to_iri
from iolanta.ensure_is_context import ensure_is_context
from iolanta.loaders.base import Loader, PyLDOptions, PyLDResponse
from iolanta.models import LDContext, LDDocument, Quad


@dataclass(frozen=True)
class SchemeChoiceLoader(Loader[URL]):
    """Try to load a file via several loaders."""

    loader_by_scheme: Dict[str, Loader[URL]]

    def __call__(self, source: str, options: PyLDOptions) -> PyLDResponse:
        """Compile document for PyLD."""
        source = URL(source)

        document = ensure_is_context(
            self.as_jsonld_document(
                source=source,
                iri=url_to_iri(source),
            ),
        )

        return {
            'document': document,
            'contextUrl': None,
            'contentType': 'application/ld+json',
            'documentUrl': str(source),
        }

    def resolve_loader_by_url(self, url: URL):
        """Find loader instance by URL."""
        try:
            return self.loader_by_scheme[url.scheme]
        except (KeyError, AttributeError):
            raise ValueError(f'Cannot find a loader for URL: {url}')

    def as_jsonld_document(
        self,
        source: URL,
        iri: Optional[URIRef] = None,
    ) -> LDDocument:
        """Represent a file as a JSON-LD document."""
        return self.resolve_loader_by_url(
            url=source,
        ).as_jsonld_document(
            source=source,
            iri=iri,
        )

    def as_quad_stream(
        self,
        source: str,
        iri: Optional[URIRef],
        root_loader: Optional[Loader[URL]] = None,
        context: Optional[LDContext] = None,
    ) -> Iterable[Quad]:
        """Convert data into a stream of RDF quads."""
        return self.resolve_loader_by_url(
            url=source,
        ).as_quad_stream(
            source=source,
            iri=iri,
            root_loader=root_loader or self,
            context=context,
        )
